@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2019 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2022 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -52,6 +52,7 @@
 #include <limits.h>
 #include <ctype.h>
 #include <setjmp.h>
+#include <math.h>
 
 #include "gettext.h"
 #define _(msgid)  gettext(msgid)
@@ -87,8 +88,6 @@ extern int errno;
 #include <wchar.h>
 #include <wctype.h>
 
-#include "mbsupport.h" /* defines stuff for DJGPP to fake MBS */
-
 #ifdef STDC_HEADERS
 #include <float.h>
 #endif
@@ -157,22 +156,13 @@ typedef int off_t;
 #define setlocale(locale, val)	/* nothing */
 #endif /* HAVE_SETLOCALE */
 
-#if HAVE_MEMCPY_ULONG
-extern char *memcpy_ulong(char *dest, const char *src, unsigned long l);
-#define memcpy memcpy_ulong
-#endif
-#if HAVE_MEMSET_ULONG
-extern void *memset_ulong(void *dest, int val, unsigned long l);
-#define memset memset_ulong
-#endif
-
 #ifdef HAVE_FWRITE_UNLOCKED
 #define fwrite	fwrite_unlocked
 #endif /* HAVE_FWRITE_UNLOCKED */
 
-#if defined(__DJGPP__) || defined(__EMX__) || defined(__MINGW32__)
+#if defined(__MINGW32__)
 #include "nonposix.h"
-#endif /* defined(__DJGPP__) || defined(__EMX__) || defined(__MINGW32__) */
+#endif /* defined(__MINGW32__) */
 
 /* use this as lintwarn("...")
    this is a hack but it gives us the right semantics */
@@ -214,6 +204,8 @@ typedef struct Regexp {
 
 #include "gawkapi.h"
 
+#include "floatmagic.h"
+
 /* Stuff for losing systems. */
 #if !defined(HAVE_STRTOD)
 extern double gawk_strtod();
@@ -259,6 +251,7 @@ typedef enum nodevals {
 	Node_var,		/* scalar variable, lnode is value */
 	Node_var_array,		/* array is ptr to elements, table_size num of eles */
 	Node_var_new,		/* newly created variable, may become an array */
+	Node_elem_new,		/* newly created array element, may become a subarray */
 	Node_param_list,	/* lnode is a variable, rnode is more list */
 	Node_func,		/* lnode is param. list, rnode is body */
 	Node_ext_func,		/* extension function, code_ptr is builtin code */
@@ -365,9 +358,10 @@ typedef struct exp_node {
 			size_t reserved;
 			struct exp_node *rn;
 			unsigned long cnt;
-			unsigned long reflags;
-#				define	CONSTANT	1
-#				define	FS_DFLT		2
+			enum reflagvals {
+				CONSTANT = 1,
+				FS_DFLT  = 2,
+			} reflags;
 		} nodep;
 
 		struct {
@@ -383,7 +377,6 @@ typedef struct exp_node {
 #endif
 			char *sp;
 			size_t slen;
-			long sref;
 			int idx;
 			wchar_t *wsp;
 			size_t wslen;
@@ -392,79 +385,85 @@ typedef struct exp_node {
 		} val;
 	} sub;
 	NODETYPE type;
-	unsigned int flags;
+	enum flagvals {
+	/* type = Node_val */
+		/*
+		 * STRING and NUMBER are mutually exclusive, except for the special
+		 * case of an uninitialized value, represented internally by
+		 * Nnull_string. They represent the type of a value as assigned.
+		 * Nnull_string has both STRING and NUMBER attributes, but all other
+		 * scalar values should have precisely one of these bits set.
+		 *
+		 * STRCUR and NUMCUR are not mutually exclusive. They represent that
+		 * the particular type of value is up to date.  For example,
+		 *
+		 * 	a = 5		# NUMBER | NUMCUR
+		 * 	b = a ""	# Adds STRCUR to a, since a string value
+		 * 			# is now available. But the type hasn't changed!
+		 *
+		 * 	a = "42"	# STRING | STRCUR
+		 * 	b = a + 0	# Adds NUMCUR to a, since numeric value
+		 * 			# is now available. But the type hasn't changed!
+		 *
+		 * USER_INPUT is the joker.  When STRING|USER_INPUT is set, it means
+		 * "this is string data, but the user may have really wanted it to be a
+		 * number. If we have to guess, like in a comparison, turn it into a
+		 * number if the string is indeed numeric."
+		 * For example,    gawk -v a=42 ....
+		 * Here, `a' gets STRING|STRCUR|USER_INPUT and then when used where
+		 * a number is needed, it gets turned into a NUMBER and STRING
+		 * is cleared. In that case, we leave the USER_INPUT in place, so
+		 * the combination NUMBER|USER_INPUT means it is a strnum a.k.a. a
+		 * "numeric string".
+		 *
+		 * WSTRCUR is for efficiency. If in a multibyte locale, and we
+		 * need to do something character based (substr, length, etc.)
+		 * we create the corresponding wide character string and store it,
+		 * and add WSTRCUR to the flags so that we don't have to do the
+		 * conversion more than once.
+		 *
+		 * The NUMINT flag may be used with a value of any type -- NUMBER,
+		 * STRING, or STRNUM. It indicates that the string representation
+		 * equals the result of sprintf("%ld", <numeric value>). So, for
+		 * example, NUMINT should NOT be set if it's a strnum or string value
+		 * where the string is " 1" or "01" or "+1" or "1.0" or "0.1E1". This
+		 * is a hint to indicate that an integer array optimization may be
+		 * used when this value appears as a subscript.
+		 *
+		 * The BOOL flag indicates that this number should be converted to True
+		 * or False by extensions that interchange data with other languages,
+		 * via JSON, XML or some other serialization mechanism.
+		 *
+		 * We hope that the rest of the flags are self-explanatory. :-)
+		 */
+		MALLOC	= 0x0001,       /* stptr can be free'd, i.e. not a field node pointing into a shared buffer */
+		STRING	= 0x0002,       /* assigned as string */
+		STRCUR	= 0x0004,       /* string value is current */
+		NUMCUR	= 0x0008,       /* numeric value is current */
+		NUMBER	= 0x0010,       /* assigned as number */
+		USER_INPUT = 0x0020,    /* user input: if NUMERIC then
+					 * a NUMBER */
+		BOOLVAL = 0x0040,	/* this is a boolean value */
+		INTLSTR	= 0x0080,       /* use localized version */
+		NUMINT	= 0x0100,       /* numeric value is an integer */
+		INTIND	= 0x0200,	/* integral value is array index;
+					 * lazy conversion to string.
+					 */
+		WSTRCUR	= 0x0400,	/* wide str value is current */
+		MPFN	= 0x0800,	/* arbitrary-precision floating-point number */
+		MPZN	= 0x01000,	/* arbitrary-precision integer */
+		NO_EXT_SET = 0x02000,	/* extension cannot set a value for this variable */
+		NULL_FIELD = 0x04000,	/* this is the null field */
 
-/* type = Node_val */
-	/*
-	 * STRING and NUMBER are mutually exclusive, except for the special
-	 * case of an uninitialized value, represented internally by
-	 * Nnull_string. They represent the type of a value as assigned.
-	 * Nnull_string has both STRING and NUMBER attributes, but all other
-	 * scalar values should have precisely one of these bits set.
-	 *
-	 * STRCUR and NUMCUR are not mutually exclusive. They represent that
-	 * the particular type of value is up to date.  For example,
-	 *
-	 * 	a = 5		# NUMBER | NUMCUR
-	 * 	b = a ""	# Adds STRCUR to a, since a string value
-	 * 			# is now available. But the type hasn't changed!
-	 *
-	 * 	a = "42"	# STRING | STRCUR
-	 * 	b = a + 0	# Adds NUMCUR to a, since numeric value
-	 * 			# is now available. But the type hasn't changed!
-	 *
-	 * USER_INPUT is the joker.  When STRING|USER_INPUT is set, it means
-	 * "this is string data, but the user may have really wanted it to be a
-	 * number. If we have to guess, like in a comparison, turn it into a
-	 * number if the string is indeed numeric."
-	 * For example,    gawk -v a=42 ....
-	 * Here, `a' gets STRING|STRCUR|USER_INPUT and then when used where
-	 * a number is needed, it gets turned into a NUMBER and STRING
-	 * is cleared. In that case, we leave the USER_INPUT in place, so
-	 * the combination NUMBER|USER_INPUT means it is a strnum a.k.a. a
-	 * "numeric string".
-	 *
-	 * WSTRCUR is for efficiency. If in a multibyte locale, and we
-	 * need to do something character based (substr, length, etc.)
-	 * we create the corresponding wide character string and store it,
-	 * and add WSTRCUR to the flags so that we don't have to do the
-	 * conversion more than once.
-	 *
-	 * The NUMINT flag may be used with a value of any type -- NUMBER,
-	 * STRING, or STRNUM. It indicates that the string representation
-	 * equals the result of sprintf("%ld", <numeric value>). So, for
-	 * example, NUMINT should NOT be set if it's a strnum or string value
-	 * where the string is " 1" or "01" or "+1" or "1.0" or "0.1E1". This
-	 * is a hint to indicate that an integer array optimization may be
-	 * used when this value appears as a subscript.
-	 *
-	 * We hope that the rest of the flags are self-explanatory. :-)
-	 */
-#		define	MALLOC	0x0001       /* stptr can be free'd, i.e. not a field node pointing into a shared buffer */
-#		define	STRING	0x0002       /* assigned as string */
-#		define	STRCUR	0x0004       /* string value is current */
-#		define	NUMCUR	0x0008       /* numeric value is current */
-#		define	NUMBER	0x0010       /* assigned as number */
-#		define	USER_INPUT 0x0020    /* user input: if NUMERIC then
-		                              * a NUMBER */
-#		define	INTLSTR	0x0040       /* use localized version */
-#		define	NUMINT	0x0080       /* numeric value is an integer */
-#		define	INTIND	0x0100	     /* integral value is array index;
-		                              * lazy conversion to string.
-		                              */
-#		define	WSTRCUR	0x0200       /* wide str value is current */
-#		define	MPFN	0x0400       /* arbitrary-precision floating-point number */
-#		define	MPZN	0x0800       /* arbitrary-precision integer */
-#		define	NO_EXT_SET 0x1000    /* extension cannot set a value for this variable */
-#		define	NULL_FIELD 0x2000    /* this is the null field */
-
-/* type = Node_var_array */
-#		define	ARRAYMAXED	0x4000       /* array is at max size */
-#		define	HALFHAT		0x8000       /* half-capacity Hashed Array Tree;
-		                                      * See cint_array.c */
-#		define	XARRAY		0x10000
-#		define	NUMCONSTSTR	0x20000	/* have string value for numeric constant */
-#		define  REGEX           0x40000 /* this is a typed regex */
+	/* type = Node_var_array */
+		ARRAYMAXED	= 0x08000,	/* array is at max size */
+		HALFHAT		= 0x010000,	/* half-capacity Hashed Array Tree;
+						 * See cint_array.c */
+		XARRAY		= 0x020000,
+		NUMCONSTSTR	= 0x040000,	/* have string value for numeric constant */
+		REGEX           = 0x080000,	/* this is a typed regex */
+	} flags;
+	long valref;
 } NODE;
 
 #define vname sub.nodep.name
@@ -501,7 +500,6 @@ typedef struct exp_node {
  */
 #define stptr	sub.val.sp
 #define stlen	sub.val.slen
-#define valref	sub.val.sref
 #define stfmt	sub.val.idx
 #define strndmode sub.val.rndmode
 #define wstptr	sub.val.wsp
@@ -715,6 +713,7 @@ typedef enum opcodeval {
 	Op_exec_count,
 	Op_breakpoint,
 	Op_lint,
+	Op_lint_plus,
 	Op_atexit,
 	Op_stop,
 
@@ -750,6 +749,16 @@ enum redirval {
 
 struct break_point;
 
+#if __DECC && __VAX
+typedef unsigned long exec_count_t;	// for exec_count
+#define EXEC_COUNT_FMT	"%lu"
+#define EXEC_COUNT_PROFILE_FMT	"%6lu"
+#else
+typedef unsigned long long exec_count_t;	// for exec_count
+#define EXEC_COUNT_FMT	"%llu"
+#define EXEC_COUNT_PROFILE_FMT	"%6llu"
+#endif
+
 typedef struct exp_instruction {
 	struct exp_instruction *nexti;
 	union {
@@ -760,6 +769,7 @@ typedef struct exp_instruction {
 					awk_value_t *result,
 					struct awk_ext_func *finfo);
 		long dl;
+		exec_count_t ldl;	// for exec_count
 		char *name;
 	} d;
 
@@ -900,7 +910,7 @@ typedef struct exp_instruction {
 
 /*------------------ pretty printing/profiling --------*/
 /* Op_exec_count */
-#define exec_count      d.dl
+#define exec_count      d.ldl
 
 /* Op_K_while */
 #define while_body      d.di
@@ -944,30 +954,33 @@ typedef struct iobuf {
 	bool valid;
 	int errcode;
 
-	int flag;
-#		define	IOP_IS_TTY	1
-#		define  IOP_AT_EOF      2
-#		define  IOP_CLOSED      4
-#		define  IOP_AT_START    8
+	enum iobuf_flags {
+		IOP_IS_TTY	= 1,
+		IOP_AT_EOF	= 2,
+		IOP_CLOSED	= 4,
+		IOP_AT_START	= 8,
+	} flag;
 } IOBUF;
 
 typedef void (*Func_ptr)(void);
 
 /* structure used to dynamically maintain a linked-list of open files/pipes */
 struct redirect {
-	unsigned int flag;
-#		define	RED_FILE	1
-#		define	RED_PIPE	2
-#		define	RED_READ	4
-#		define	RED_WRITE	8
-#		define	RED_APPEND	16
-#		define	RED_NOBUF	32
-#		define	RED_USED	64	/* closed temporarily to reuse fd */
-#		define	RED_EOF		128
-#		define	RED_TWOWAY	256
-#		define	RED_PTY		512
-#		define	RED_SOCKET	1024
-#		define	RED_TCP		2048
+	enum redirect_flags {
+		RED_NONE	= 0,
+		RED_FILE	= 1,
+		RED_PIPE	= 2,
+		RED_READ	= 4,
+		RED_WRITE	= 8,
+		RED_APPEND	= 16,
+		RED_NOBUF	= 32,
+		RED_USED	= 64,	/* closed temporarily to reuse fd */
+		RED_EOF		= 128,
+		RED_TWOWAY	= 256,
+		RED_PTY		= 512,
+		RED_SOCKET	= 1024,
+		RED_TCP		= 2048,
+	} flag;
 	char *value;
 	FILE *ifp;	/* input fp, needed for PIPES_SIMULATED */
 	IOBUF *iop;
@@ -978,14 +991,15 @@ struct redirect {
 	const char *mode;
 	awk_output_buf_t output;
 };
+typedef enum redirect_flags redirect_flags_t;
 
 /* values for BINMODE, used as bit flags */
 
 enum binmode_values {
-	TEXT_TRANSLATE = 0,	/* usual \r\n ---> \n translation */
-	BINMODE_INPUT = 1,	/* no translation for input files */
-	BINMODE_OUTPUT = 2,	/* no translation for output files */
-	BINMODE_BOTH = 3	/* no translation for either */
+	TEXT_TRANSLATE	= 0,	/* usual \r\n ---> \n translation */
+	BINMODE_INPUT	= 1,	/* no translation for input files */
+	BINMODE_OUTPUT	= 2,	/* no translation for output files */
+	BINMODE_BOTH	= 3	/* no translation for either */
 };
 
 /*
@@ -1061,13 +1075,16 @@ struct block_item {
 struct block_header {
 	struct block_item *freep;
 	size_t size;
+	const char *name;
+	long highwater;
+#ifdef MEMDEBUG
+	long active;
+#endif
 };
 
 enum block_id {
 	BLOCK_NODE = 0,
 	BLOCK_BUCKET,
-	BLOCK_MPFR,
-	BLOCK_MPZ,
 	BLOCK_MAX	/* count */
 };
 
@@ -1098,7 +1115,7 @@ extern int OFSlen;
 extern char *ORS;
 extern int ORSlen;
 extern char *OFMT;
-extern char *CONVFMT;
+extern const char *CONVFMT;
 extern int CONVFMTidx;
 extern int OFMTidx;
 #ifdef HAVE_MPFR
@@ -1134,11 +1151,12 @@ extern NODE *success_node;
 extern struct block_header nextfree[];
 extern bool field0_valid;
 
-extern int do_flags;
+extern bool do_itrace;	/* separate so can poke from a debugger */
 
 extern SRCFILE *srcfiles; /* source files */
 
-enum do_flag_values {
+extern enum do_flag_values {
+	DO_FLAG_NONE       = 0x00000,
 	DO_LINT_INVALID	   = 0x00001,	/* only warn about invalid */
 	DO_LINT_EXTENSIONS = 0x00002,	/* warn about gawk extensions */
 	DO_LINT_ALL	   = 0x00004,	/* warn about all things */
@@ -1154,8 +1172,8 @@ enum do_flag_values {
 	DO_SANDBOX	   = 0x01000,	/* sandbox mode - disable 'system' function & redirections */
 	DO_PROFILE	   = 0x02000,	/* profile the program */
 	DO_DEBUG	   = 0x04000,	/* debug the program */
-	DO_MPFR		   = 0x08000	/* arbitrary-precision floating-point math */
-};
+	DO_MPFR		   = 0x08000,	/* arbitrary-precision floating-point math */
+} do_flags;
 
 #define do_traditional      (do_flags & DO_TRADITIONAL)
 #define do_posix            (do_flags & DO_POSIX)
@@ -1173,10 +1191,12 @@ enum do_flag_values {
 extern bool do_optimize;
 extern int use_lc_numeric;
 extern int exit_val;
+extern bool using_persistent_malloc;
 
 #ifdef NO_LINT
 #define do_lint 0
 #define do_lint_old 0
+#define do_lint_extensions 0
 #else
 #define do_lint             (do_flags & (DO_LINT_INVALID|DO_LINT_ALL))
 #define do_lint_old         (do_flags & DO_LINT_OLD)
@@ -1206,10 +1226,10 @@ extern bool do_ieee_fmt;	/* emulate IEEE 754 floating-point format */
 extern const char *myname;
 extern const char def_strftime_format[];
 
-extern char quote;
-extern char *defpath;
-extern char *deflibpath;
-extern char envsep;
+extern const char quote;
+extern const char *defpath;
+extern const char *deflibpath;
+extern const char envsep;
 
 extern char casetable[];	/* for case-independent regexp matching */
 
@@ -1261,8 +1281,11 @@ static inline void
 DEREF(NODE *r)
 {
 	assert(r->valref > 0);
-	if (--r->valref == 0)
-		r_unref(r);
+#ifndef GAWKDEBUG
+	if (--r->valref > 0)
+		return;
+#endif
+	r_unref(r);
 }
 
 #define POP_NUMBER() force_number(POP_SCALAR())
@@ -1297,7 +1320,7 @@ DEREF(NODE *r)
 
 #define get_number_uj(n)	numtype_choose((n), mpfr_get_uj((n)->mpg_numbr, ROUND_MODE), (uintmax_t) mpz_get_d((n)->mpg_i), (uintmax_t) (n)->numbr)
 
-#define iszero(n)		numtype_choose((n), mpfr_zero_p((n)->mpg_numbr), (mpz_sgn((n)->mpg_i) == 0), ((n)->numbr == 0.0))
+#define is_zero(n)		numtype_choose((n), mpfr_zero_p((n)->mpg_numbr), (mpz_sgn((n)->mpg_i) == 0), ((n)->numbr == 0.0))
 
 #define IEEE_FMT(r, t)		(void) (do_ieee_fmt && format_ieee(r, t))
 
@@ -1315,7 +1338,7 @@ DEREF(NODE *r)
 #define is_mpg_number(n)	0
 #define is_mpg_float(n)		0
 #define is_mpg_integer(n)	0
-#define iszero(n)		((n)->numbr == 0.0)
+#define is_zero(n)		((n)->numbr == 0.0)
 #endif
 
 #define var_uninitialized(n)	((n)->var_value == Nnull_string)
@@ -1323,12 +1346,23 @@ DEREF(NODE *r)
 #define get_lhs(n, r)	 (n)->type == Node_var && ! var_uninitialized(n) ? \
 				&((n)->var_value) : r_get_lhs((n), (r))
 
+#ifdef MEMDEBUG
+
+extern void *r_getblock(int id);
+extern void r_freeblock(void *, int id);
+#define getblock(p, id, ty)	(void) (p = (ty) r_getblock(id))
+#define freeblock(p, id)	(void) (r_freeblock(p, id))
+
+#else /* MEMDEBUG */
+
 #define getblock(p, id, ty)  (void) ((p = (ty) nextfree[id].freep) ? \
 			(ty) (nextfree[id].freep = ((struct block_item *) p)->freep) \
 			: (p = (ty) more_blocks(id)))
 #define freeblock(p, id)	 (void) (((struct block_item *) p)->freep = nextfree[id].freep, \
 					nextfree[id].freep = (struct block_item *) p)
 
+#endif /* MEMDEBUG */
+
 #define getnode(n)	getblock(n, BLOCK_NODE, NODE *)
 #define freenode(n)	freeblock(n, BLOCK_NODE)
 
@@ -1342,8 +1376,8 @@ DEREF(NODE *r)
 #define		ALREADY_MALLOCED	2
 #define		ELIDE_BACK_NL		4
 
-#define	cant_happen()	r_fatal("internal error line %d, file: %s", \
-				__LINE__, __FILE__)
+#define	cant_happen(format, ...)	r_fatal("internal error: file %s, line %d: " format, \
+				__FILE__, __LINE__, __VA_ARGS__)
 
 #define	emalloc(var,ty,x,str)	(void) (var = (ty) emalloc_real((size_t)(x), str, #var, __FILE__, __LINE__))
 #define	ezalloc(var,ty,x,str)	(void) (var = (ty) ezalloc_real((size_t)(x), str, #var, __FILE__, __LINE__))
@@ -1405,6 +1439,7 @@ extern NODE *do_asorti(int nargs);
 extern unsigned long (*hash)(const char *s, size_t len, unsigned long hsize, size_t *code);
 extern void init_env_array(NODE *env_node);
 extern void init_argv_array(NODE *argv_node, NODE *shadow_node);
+extern NODE *new_array_element(void);
 /* awkgram.c */
 extern NODE *variable(int location, char *name, NODETYPE type);
 extern int parse_program(INSTRUCTION **pcode, bool from_eval);
@@ -1427,9 +1462,10 @@ extern bool is_alpha(int c);
 extern bool is_alnum(int c);
 extern bool is_letter(int c);
 extern bool is_identchar(int c);
-extern NODE *make_regnode(int type, NODE *exp);
+extern NODE *make_regnode(NODETYPE type, NODE *exp);
 extern bool validate_qualified_name(char *token);
 /* builtin.c */
+extern void efflush(FILE *fp, const char *from, struct redirect *rp);
 extern double double_to_int(double d);
 extern NODE *do_exp(int nargs);
 extern NODE *do_fflush(int nargs);
@@ -1478,6 +1514,10 @@ extern NODE *do_typeof(int nargs);
 extern int strncasecmpmbs(const unsigned char *,
 			  const unsigned char *, size_t);
 extern int sanitize_exit_status(int status);
+extern void check_symtab_functab(NODE *dest, const char *fname, const char *msg);
+extern NODE *do_mkbool(int nargs);
+extern void check_exact_args(int nargs, const char *fname, int count);
+extern void check_args_min_max(int nargs, const char *fname, int min, int max);
 /* debug.c */
 extern void init_debug(void);
 extern int debug_prog(INSTRUCTION *pc);
@@ -1514,6 +1554,7 @@ extern STACK_ITEM *grow_stack(void);
 extern void dump_fcall_stack(FILE *fp);
 extern int register_exec_hook(Func_pre_exec preh, Func_post_exec posth);
 extern NODE **r_get_field(NODE *n, Func_ptr *assign, bool reference);
+extern NODE *elem_new_to_scalar(NODE *n);
 /* ext.c */
 extern NODE *do_ext(int nargs);
 void load_ext(const char *lib_name);	/* temporary */
@@ -1551,6 +1592,15 @@ typedef enum {
 extern field_sep_type current_field_sep(void);
 extern const char *current_field_sep_str(void);
 
+typedef enum {
+	SCALAR_EQ,
+	SCALAR_NEQ,
+	SCALAR_LT,
+	SCALAR_LE,
+	SCALAR_GT,
+	SCALAR_GE,
+} scalar_cmp_t;
+
 /* gawkapi.c: */
 extern gawk_api_t api_impl;
 extern void init_ext_api(void);
@@ -1561,7 +1611,7 @@ extern void print_ext_versions(void);
 extern void free_api_string_copies(void);
 
 /* gawkmisc.c */
-extern char *gawk_name(const char *filespec);
+extern const char *gawk_name(const char *filespec);
 extern void os_arg_fixup(int *argcp, char ***argvp);
 extern int os_devopen(const char *name, int flag);
 extern void os_close_on_exec(int fd, const char *name, const char *what, const char *dir);
@@ -1571,6 +1621,7 @@ extern int os_isreadable(const awk_input_buf_t *iobuf, bool *isdir);
 extern int os_is_setuid(void);
 extern int os_setbinmode(int fd, int mode);
 extern void os_restore_mode(int fd);
+extern void os_maybe_set_errno(void);
 extern size_t optimal_bufsize(int fd, struct stat *sbuf);
 extern int ispath(const char *file);
 extern int isdirpunct(int c);
@@ -1607,6 +1658,7 @@ extern bool is_non_fatal_redirect(const char *str, size_t len);
 extern void ignore_sigpipe(void);
 extern void set_sigpipe_to_default(void);
 extern bool non_fatal_flush_std_file(FILE *fp);
+extern size_t gawk_fwrite(const void *buf, size_t size, size_t count, FILE *fp, void *opaque);
 
 /* main.c */
 extern int arg_assign(char *arg, bool initing);
@@ -1649,6 +1701,10 @@ extern void cleanup_mpfr(void);
 extern NODE *mpg_node(unsigned int);
 extern const char *mpg_fmt(const char *, ...);
 extern int mpg_strtoui(mpz_ptr, char *, size_t, char **, int);
+extern void mpg_zero(NODE *n);
+extern void *mpfr_mem_alloc(size_t alloc_size);
+extern void *mpfr_mem_realloc(void *ptr, size_t old_size, size_t new_size);
+extern void mpfr_mem_free(void *ptr, size_t size);
 #endif
 /* msg.c */
 extern void gawk_exit(int status);
@@ -1679,6 +1735,7 @@ extern NODE *r_force_number(NODE *n);
 extern NODE *r_format_val(const char *format, int index, NODE *s);
 extern NODE *r_dupnode(NODE *n);
 extern NODE *make_str_node(const char *s, size_t len, int flags);
+extern NODE *make_bool_node(bool value);
 extern NODE *make_typed_regex(const char *re, size_t len);
 extern void *more_blocks(int id);
 extern int parse_escape(const char **string_ptr);
@@ -1697,6 +1754,7 @@ extern void init_btowc_cache();
 #define is_valid_character(b)	(btowc_cache[(b)&0xFF] != WEOF)
 extern bool out_of_range(NODE *n);
 extern char *format_nan_inf(NODE *n, char format);
+extern bool is_ieee_magic_val(const char *val);
 /* re.c */
 extern Regexp *make_regexp(const char *s, size_t len, bool ignorecase, bool dfa, bool canfatal);
 extern int research(Regexp *rp, char *str, int start, size_t len, int flags);
@@ -1791,7 +1849,8 @@ POP_ARRAY(bool check_for_untyped)
 	NODE *t = POP();
 	static bool warned = false;
 
-	if (do_lint && ! warned && check_for_untyped && t->type == Node_var_new) {
+	if (do_lint && ! warned && check_for_untyped
+	    && (t->type == Node_var_new || t->type == Node_elem_new)) {
 		warned = true;
 		lintwarn(_("behavior of `for' loop on untyped variable is not defined by POSIX"));
 	}
@@ -1818,6 +1877,8 @@ POP_SCALAR()
 
 	if (t->type == Node_var_array)
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t));
+	else if (t->type == Node_elem_new)
+		t = elem_new_to_scalar(t);
 
 	return t;
 }
@@ -1831,6 +1892,10 @@ TOP_SCALAR()
 
 	if (t->type == Node_var_array)
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t));
+	else if (t->type == Node_elem_new) {
+		t = elem_new_to_scalar(t);	// fix it up
+		REPLACE(t);			// put it back on the stack
+	}
 
 	return t;
 }
@@ -1884,6 +1949,13 @@ dupnode(NODE *n)
 static inline NODE *
 force_string_fmt(NODE *s, const char *fmtstr, int fmtidx)
 {
+	if (s->type == Node_elem_new) {
+		s->type = Node_val;
+		s->flags &= ~NUMBER;
+
+		return s;
+	}
+
 	if ((s->flags & STRCUR) != 0
 		&& (s->stfmt == STFMT_UNUSED || (s->stfmt == fmtidx
 #ifdef HAVE_MPFR
@@ -1909,6 +1981,7 @@ force_string_fmt(NODE *s, const char *fmtstr, int fmtidx)
 static inline void
 unref(NODE *r)
 {
+	assert(r == NULL || r->valref > 0);
 	if (r != NULL && --r->valref <= 0)
 		r_unref(r);
 }
@@ -1959,7 +2032,7 @@ boolval(NODE *t)
 {
 	(void) fixtype(t);
 	if ((t->flags & NUMBER) != 0)
-		return ! iszero(t);
+		return ! is_zero(t);
 	return (t->stlen > 0);
 }
 
@@ -1975,7 +2048,7 @@ emalloc_real(size_t count, const char *where, const char *var, const char *file,
 
 	ret = (void *) malloc(count);
 	if (ret == NULL)
-		fatal(_("%s:%d:%s: %s: can't allocate %ld bytes of memory (%s)"),
+		fatal(_("%s:%d:%s: %s: cannot allocate %ld bytes of memory: %s"),
 			file, line, where, var, (long) count, strerror(errno));
 
 	return ret;
@@ -1993,7 +2066,7 @@ ezalloc_real(size_t count, const char *where, const char *var, const char *file,
 
 	ret = (void *) calloc(1, count);
 	if (ret == NULL)
-		fatal(_("%s:%d:%s: %s: can't allocate %ld bytes of memory (%s)"),
+		fatal(_("%s:%d:%s: %s: cannot allocate %ld bytes of memory: %s"),
 			file, line, where, var, (long) count, strerror(errno));
 
 	return ret;
@@ -2011,7 +2084,7 @@ erealloc_real(void *ptr, size_t count, const char *where, const char *var, const
 
 	ret = (void *) realloc(ptr, count);
 	if (ret == NULL)
-		fatal(_("%s:%d:%s: %s: can't reallocate %ld bytes of memory (%s)"),
+		fatal(_("%s:%d:%s: %s: cannot reallocate %ld bytes of memory: %s"),
 			file, line, where, var, (long) count, strerror(errno));
 
 	return ret;
