@@ -3,7 +3,7 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2021,
+ * Copyright (C) 1986, 1988, 1989, 1991-2022,
  * the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
@@ -67,6 +67,7 @@ r_interpret(INSTRUCTION *code)
 	Regexp *rp;
 	NODE *set_array = NULL;	/* array with a post-assignment routine */
 	NODE *set_idx = NULL;	/* the index of the array element */
+	bool in_indirect_call = false;
 
 
 /* array subscript */
@@ -223,6 +224,20 @@ uninitialized_scalar:
 								save_symbol->vname);
 				if (op != Op_push_arg_untyped)
 					m = dupnode(Nnull_string);
+				UPREF(m);
+				PUSH(m);
+				break;
+
+			case Node_elem_new:
+				if (op != Op_push_arg_untyped) {
+					/* convert untyped to scalar */
+					m = elem_new_to_scalar(m);
+				}
+				if (do_lint)
+					lintwarn(isparam ?
+						_("reference to uninitialized argument `%s'") :
+						_("reference to uninitialized variable `%s'"),
+								save_symbol->vname);
 				PUSH(m);
 				break;
 
@@ -235,7 +250,7 @@ uninitialized_scalar:
 				break;
 
 			default:
-				cant_happen();
+				cant_happen("unexpected parameter type %s", nodetype2str(m->type));
 			}
 		}
 			break;
@@ -319,7 +334,7 @@ uninitialized_scalar:
 				}
 			}
 
-			if (r->type == Node_val)
+			if (r->type == Node_val || r->type == Node_elem_new)
 				UPREF(r);
 			PUSH(r);
 			break;
@@ -328,12 +343,35 @@ uninitialized_scalar:
 			t2 = mk_sub(pc->sub_count);
 			t1 = POP_ARRAY(false);
 			r = in_array(t1, t2);
+
+			if (r == NULL) {
+				t2 = force_string(t2);
+
+				if (t1 == func_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"FUNCTAB", (int) t2->stlen, t2->stptr);
+				} else if (t1 == symbol_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"SYMTAB", (int) t2->stlen, t2->stptr);
+				} else if (do_lint) {
+					lintwarn(_("reference to uninitialized element `%s[\"%.*s\"]'"),
+						array_vname(t1), (int) t2->stlen, t2->stptr);
+					if (t2->stlen == 0)
+						lintwarn(_("subscript of array `%s' is null string"), array_vname(t1));
+				}
+			}
+
 			if (r == NULL) {
 				r = make_array();
 				r->parent_array = t1;
 				t2 = force_string(t2);
 				r->vname = estrdup(t2->stptr, t2->stlen);	/* the subscript in parent array */
 				assoc_set(t1, t2, r);
+			} else if (r->type == Node_elem_new) {
+				r = force_array(r, false);
+				r->parent_array = t1;
+				t2 = force_string(t2);
+				r->vname = estrdup(t2->stptr, t2->stlen);	/* the subscript in parent array */
 			} else if (r->type != Node_var_array) {
 				t2 = force_string(t2);
 				fatal(_("attempt to use scalar `%s[\"%.*s\"]' as an array"),
@@ -370,7 +408,7 @@ uninitialized_scalar:
 			 * be stored in SYMTAB:
 			 * 	1. Variables that don"t yet have a value (Node_var_new)
 			 * 	2. Variables that have a value (Node_var)
-			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_value)
+			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_val)
 			 * For 1, since we are giving it a value, we have to change the type to Node_var.
 			 * For 1 and 2, we have to step through the Node_var to get to the value.
 			 * For 3, we fatal out. This avoids confusion on things like
@@ -426,7 +464,7 @@ uninitialized_scalar:
 					break;
 
 				default:
-					cant_happen();
+					cant_happen("unexpected lint type value %d", (int) pc->lint_type);
 				}
 			}
 			break;
@@ -493,37 +531,37 @@ uninitialized_scalar:
 			break;
 
 		case Op_equal:
-			r = node_Boolean[cmp_scalars(SCALAR_EQ_NEQ) == 0];
+			r = node_Boolean[cmp_scalars(SCALAR_EQ)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_notequal:
-			r = node_Boolean[cmp_scalars(SCALAR_EQ_NEQ) != 0];
+			r = node_Boolean[cmp_scalars(SCALAR_NEQ)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_less:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) < 0];
+			r = node_Boolean[cmp_scalars(SCALAR_LT)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_greater:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) > 0];
+			r = node_Boolean[cmp_scalars(SCALAR_GT)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_leq:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) <= 0];
+			r = node_Boolean[cmp_scalars(SCALAR_LE)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_geq:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) >= 0];
+			r = node_Boolean[cmp_scalars(SCALAR_GE)];
 			UPREF(r);
 			REPLACE(r);
 			break;
@@ -694,7 +732,7 @@ mod:
 			 * SYMTAB is a little more messy.  Three possibilities for SYMTAB:
 			 * 	1. Variables that don"t yet have a value (Node_var_new)
 			 * 	2. Variables that have a value (Node_var)
-			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_value)
+			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_val)
 			 * For 1, since we are giving it a value, we have to change the type to Node_var.
 			 * For 1 and 2, we have to step through the Node_var to get to the value.
 			 * For 3, we fatal out. This avoids confusion on things like
@@ -825,6 +863,10 @@ mod:
 			lhs = POP_ADDRESS();
 			r = TOP_SCALAR();
 			unref(*lhs);
+			if (r->type == Node_elem_new) {
+				DEREF(r);
+				r = dupnode(Nnull_string);
+			}
 			UPREF(r);
 			UNFIELD(*lhs, r);
 			REPLACE(r);
@@ -1055,10 +1097,18 @@ arrayfor:
 			(void) POP_CODE();
 			while (arg_count-- > 0) {
 				t1 = POP();
-				if (t1->type == Node_val)
+				if (t1->type == Node_val || t1->type == Node_elem_new)
 					DEREF(t1);
 			}
 			free_api_string_copies();
+
+			if (in_indirect_call) {
+				// pop function name off the stack
+				NODE *fname = POP();
+				DEREF(fname);
+				in_indirect_call = false;
+			}
+
 			PUSH(r);
 		}
 			break;
@@ -1132,6 +1182,7 @@ match_re:
 			NODE *f = NULL;
 			int arg_count;
 			char save;
+			NODE *function_name;
 
 			arg_count = (pc + 1)->expr_count;
 			t1 = PEEK(arg_count);	/* indirect var */
@@ -1174,6 +1225,12 @@ match_re:
 					r = the_func(arg_count);
 				str_restore(t1, save);
 
+				// Normally, setup_frame() handles getting rid of the
+				// function name.  Since we have called the builtin directly,
+				// we have to manually do this here.
+				function_name = POP();
+				DEREF(function_name);
+
 				PUSH(r);
 				break;
 			} else if (f->type != Node_func) {
@@ -1195,6 +1252,7 @@ match_re:
 					npc[1] = pc[1];
 					npc[1].func_name = fname;	/* name of the builtin */
 					npc[1].c_function = bc->c_function;
+					in_indirect_call = true;
 					ni = npc;
 					JUMPTO(ni);
 				} else
@@ -1243,7 +1301,7 @@ match_re:
 		}
 
 		case Op_K_return_from_eval:
-			cant_happen();
+			cant_happen("unexpected opcode %s", opcode2str(op));
 			break;
 
 		case Op_K_return:
